"""
Exceptions used in this library
"""


class ClientException(Exception):
    """
    Base Client Exception for module
    """
    pass


class InvalidReturnFieldException(ClientException):
    """
    Exception for invalid return fields
    """
    pass


class InvalidSortFieldException(ClientException):
    """
    Exception for invalid sort fields
    """
    pass


class InvalidFilterFieldException(ClientException):
    """
    Exception for invalid filter fields
    """
    pass


class InvalidResponseException(ClientException):
    """
    Exception thrown when receiving an invalid response from selected resource
    """
    pass


class BadRequestException(ClientException):
    """
    Exception thrown when attempting to send a bad request
    """
    pass



"""
The response types and factories for PyBomb
"""
from collections import namedtuple


class Response(
    namedtuple(
        'Response', ('uri', 'num_page_results', 'num_total_results', 'results')
    )
):
    """
    An API response
    """

    @classmethod
    def from_response_data(cls, response_data):
        """
        Response factory

        :param response_data: requests.models.Response
        :return: pybomb.clients.Response
        """

        response_json = response_data.json()

        return cls(
            response_data.url,
            response_json['number_of_page_results'],
            response_json['number_of_total_results'],
            response_json['results']
        )



"""
Client for the Game resource of GiantBomb
https://www.giantbomb.com/api/documentation#toc-0-16
"""
from pybomb.clients.base_client import BaseClient, ResponseParam


class GameClient(BaseClient):
    """
    Client for the 'game' API resource
    """
    RESOURCE_NAME = 'game'

    RESPONSE_FIELD_MAP = {
        'aliases': ResponseParam(True, False),
        'api_detail_url': ResponseParam(True, False),
        'characters': ResponseParam(True, False),
        'concepts': ResponseParam(True, False),
        'date_added': ResponseParam(True, False),
        'date_last_updated': ResponseParam(True, False),
        'deck': ResponseParam(True, False),
        'description': ResponseParam(True, False),
        'developers': ResponseParam(True, False),
        'expected_release_day': ResponseParam(True, False),
        'expected_release_month': ResponseParam(True, False),
        'expected_release_quarter': ResponseParam(True, False),
        'expected_release_year': ResponseParam(True, False),
        'first_appearance_characters': ResponseParam(True, False),
        'first_appearance_concepts': ResponseParam(True, False),
        'first_appearance_locations': ResponseParam(True, False),
        'first_appearance_objects': ResponseParam(True, False),
        'first_appearance_people': ResponseParam(True, False),
        'franchises': ResponseParam(True, False),
        'genres': ResponseParam(True, False),
        'id': ResponseParam(True, False),
        'image': ResponseParam(True, False),
        'images': ResponseParam(True, False),
        'killed_characters': ResponseParam(True, False),
        'locations': ResponseParam(True, False),
        'name': ResponseParam(True, False),
        'number_of_user_reviews': ResponseParam(True, False),
        'objects': ResponseParam(True, False),
        'original_game_rating': ResponseParam(False, True),
        'original_release_date': ResponseParam(True, False),
        'people': ResponseParam(True, False),
        'platforms': ResponseParam(True, False),
        'publishers': ResponseParam(True, False),
        'releases': ResponseParam(True, False),
        'reviews': ResponseParam(True, False),
        'similar_games': ResponseParam(True, False),
        'site_detail_url': ResponseParam(True, False),
        'themes': ResponseParam(True, False),
        'videos': ResponseParam(True, False),
    }

    def fetch(self, id_, return_fields=None):
        """
        Wrapper for fetching details of game by ID

        :param id_: int
        :param return_fields: tuple

        :return: pybomb.clients.Response
        """
        game_params = {'id': id_}

        if return_fields is not None:
            self._validate_return_fields(return_fields)
            field_list = ','.join(return_fields)

            game_params['field_list'] = field_list

        response = self._query(game_params, direct=True)

        return response



"""
A library of clients for the main resources of the GiantBomb API
http://www.giantbomb.com/api/documentation#toc-0-1
"""
from pybomb.clients.games_client import GamesClient
from pybomb.clients.game_client import GameClient

# Platform ID's
PS1 = 22
PS2 = 19
PS3 = 35
PS4 = 146

XBOX = 32
XBOX_360 = 20
XBOX_ONE = 145

MAC = 17
PC = 94



import pytest
from mock import MagicMock
from requests.models import Response as RequestsResponse

from pybomb.response import Response


class TestResponse:
    @pytest.fixture
    def mock_response(self):
        mock_response = MagicMock(RequestsResponse)
        mock_response.url = 'https://fake.com'

        mock_response.json.return_value = {
            'status_code': 1,
            'number_of_page_results': 1,
            'number_of_total_results': 1,
            'results': [{'id': 1, 'description': 'Great Game'}],
        }

        return mock_response

    def test_response_factory(self, mock_response):
        res = Response.from_response_data(mock_response)

        assert res.uri == mock_response.url

        mock_res_json = mock_response.json()
        assert res.num_page_results == mock_res_json['number_of_page_results']
        assert res.results == mock_res_json['results']
        assert res.num_total_results == (
            mock_res_json['number_of_total_results']
        )




"""
Client for the Games resource of GiantBomb
https://www.giantbomb.com/api/documentation#toc-0-17
"""
from pybomb.clients.base_client import BaseClient, ResponseParam


class GamesClient(BaseClient):
    """
    Client for the 'games' API resource
    """
    RESOURCE_NAME = 'games'

    RESPONSE_FIELD_MAP = {
        'aliases': ResponseParam(True, False),
        'api_detail_url': ResponseParam(False, False),
        'date_added': ResponseParam(True, True),
        'date_last_updated': ResponseParam(True, True),
        'deck': ResponseParam(False, False),
        'description': ResponseParam(False, False),
        'expected_release_month': ResponseParam(True, False),
        'expected_release_quarter': ResponseParam(True, False),
        'expected_release_year': ResponseParam(True, False),
        'id': ResponseParam(True, True),
        'image': ResponseParam(False, False),
        'name': ResponseParam(True, True),
        'number_of_user_reviews': ResponseParam(True, True),
        'original_game_rating': ResponseParam(False, True),
        'original_release_date': ResponseParam(True, True),
        'platforms': ResponseParam(True, False),
        'site_detail_url': ResponseParam(False, False),
    }

    def search(
        self, filter_by, return_fields=None, sort_by=None, desc=True,
        limit=None, offset=None
    ):
        """
        Full search of games resource, supporting all search fields
        available in API
        http://www.giantbomb.com/api/documentation#toc-0-17

        :param filter_by: dict
        :param return_fields: tuple
        :param sort_by: string
        :param desc: bool
        :param limit: int
        :param offset: int
        :return: pybomb.clients.Response
        """
        self._validate_filter_fields(filter_by)
        search_filter = self._create_search_filter(filter_by)

        search_params = {'filter': search_filter}

        if return_fields is not None:
            self._validate_return_fields(return_fields)
            field_list = ','.join(return_fields)

            search_params['field_list'] = field_list

        if sort_by is not None:
            self._validate_sort_field(sort_by)

            if desc:
                direction = self.SORT_ORDER_DESCENDING
            else:
                direction = self.SORT_ORDER_ASCENDING

            search_params['sort'] = '{0}:{1}'.format(sort_by, direction)

        if limit is not None:
            search_params['limit'] = int(limit)

        if offset is not None:
            search_params['offset'] = int(offset)

        response = self._query(search_params)

        return response

    def quick_search(self, name, platform=None, sort_by=None, desc=True):
        """
        Quick search method that allows you to search for a game using only the
        title and the platform

        :param name: string
        :param platform: int
        :param sort_by: string
        :param desc: bool
        :return: pybomb.clients.Response
        """
        if platform is None:
            query_filter = 'name:{0}'.format(name)
        else:
            query_filter = 'name:{0},platforms:{1}'.format(name, platform)

        search_params = {'filter': query_filter}

        if sort_by is not None:
            self._validate_sort_field(sort_by)

            if desc:
                direction = self.SORT_ORDER_DESCENDING
            else:
                direction = self.SORT_ORDER_ASCENDING

            search_params['sort'] = '{0}:{1}'.format(sort_by, direction)

        response = self._query(search_params)

        return response



"""
Base client to extend to create clients for endpoints of the GiantBomb API
"""
import pkg_resources
from collections import namedtuple

from requests import get
from requests.exceptions import HTTPError


from pybomb.exceptions import (
    InvalidReturnFieldException, BadRequestException,
    InvalidFilterFieldException, InvalidResponseException,
    InvalidSortFieldException
)
from pybomb.response import Response


ResponseParam = namedtuple('ResponseParam', ('is_filter', 'is_sort'))


class BaseClient(object):
    """
    Base class for API resource clients
    """
    URI_BASE = 'http://www.giantbomb.com/api/'
    RESPONSE_FORMAT_JSON = 'json'
    RESPONSE_FORMAT_XML = 'xml'
    RESPONSE_FIELD_MAP = None
    RESPONSE_STATUS_OK = 1

    RESOURCE_NAME = None

    SORT_ORDER_ASCENDING = 'asc'
    SORT_ORDER_DESCENDING = 'desc'

    def __init__(self, api_key, default_format=RESPONSE_FORMAT_JSON):
        """
        :param api_key: string
        :param default_format: string
        """
        self._api_key = api_key
        self._default_format = default_format
        self._headers = {'User-Agent': 'Pybomb {0}'.format(
            pkg_resources.require("pybomb")[0].version
        )}

    @property
    def api_key(self):
        """
        Giant Bomb API key

        :return: string
        """
        return self._api_key

    @api_key.setter
    def api_key(self, api_key):
        """
        :param api_key: string
        """
        self._api_key = api_key

    @property
    def default_format(self):
        """
        Default API response type

        :return: string
        """
        return self._default_format

    def _validate_return_fields(self, return_fields):
        """
        :param return_fields: tuple
        :raises: pybomb.exceptions.InvalidReturnFieldException
        """
        for return_field in return_fields:
            if return_field not in self.RESPONSE_FIELD_MAP:
                raise InvalidReturnFieldException(
                    '"{0}" is an invalid return field'.format(return_field)
                )

    def _validate_sort_field(self, sort_by):
        """
        :param sort_by: string
        :raises: pybomb.exceptions.InvalidSortFieldException
        """
        if (sort_by not in self.RESPONSE_FIELD_MAP or
                not self.RESPONSE_FIELD_MAP[sort_by].is_sort):
            raise InvalidSortFieldException(
                '"{0}" is an invalid sort field'.format(sort_by)
            )

    def _validate_filter_fields(self, filter_by):
        """
        :param filter_by: dict
        :raises: pybomb.exceptions.InvalidFilterFieldException
        """
        for filter_field in filter_by:
            if (filter_field not in self.RESPONSE_FIELD_MAP or
                    not self.RESPONSE_FIELD_MAP[filter_field].is_filter):
                raise InvalidFilterFieldException(
                    '"{0}" is an invalid filter field'.format(filter_field)
                )

    @staticmethod
    def _create_search_filter(filter_by):
        """
        :param filter_by:
        :return: dict
        """
        return ','.join(
            ['{0}:{1}'.format(key, value) for
             key, value in filter_by.items() if value is not None]
        )

    def _query(self, params, direct=False):
        """
        :param params: dict
        :return: pybomb.clients.response
        """
        params['api_key'] = self._api_key

        if 'format' not in params:
            params['format'] = self._default_format

        response = self._query_api(params, direct)
        self._validate_response(response)

        return Response.from_response_data(response)

    def _query_api(self, params, direct=False):
        """
        :param params: dict
        :return: requests.models.Response
        """
        if not direct:
            return get(
                self.URI_BASE + self.RESOURCE_NAME,
                params=params,
                headers=self._headers
            )

        id = params.pop('id')
        return get(
            self.URI_BASE + self.RESOURCE_NAME + '/{0}'.format(id),
            params=params,
            headers=self._headers
        )

    def _validate_response(self, response):
        """
        :param response: requests.models.Response
        :raises: pybomb.exceptions.InvalidResponseException
        :raises: pybomb.exceptions.BadRequestException
        """
        try:
            response.raise_for_status()
        except HTTPError as http_error:
            raise BadRequestException(str(http_error))

        response_data = response.json()
        if response_data['status_code'] != self.RESPONSE_STATUS_OK:
            raise InvalidResponseException(
                'Response code {0}: {1}'.format(
                    response_data['status_code'],
                    response_data['error']
                )
            )



from codecs import open
import os
from setuptools import setup, find_packages

version = '0.2.0'
download_url = (
    'https://github.com/steveYeah/PyBomb/archive/v{0}.tar.gz'.format(version)
)

here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, 'README.rst'), 'r', 'utf-8') as handle:
    readme = handle.read()

setup(
    name='pybomb',
    description='Clients for the Giant Bomb API',
    packages=find_packages(exclude=['test', 'test.*']),
    version=version,
    long_description=readme,
    license="MIT",
    author='Steve Hutchins',
    author_email='hutchinsteve@gmail.com',
    url='https://github.com/steveYeah/PyBomb',
    download_url=download_url,
    keywords=(
        'giantbomb'
        'giant',
        'bomb',
        'game',
        'api',
        'client'
    ),
    install_requires=(
        'requests',
    ),
    extras_require={
        'dev': (
            'pytest',
            'pycodestyle',
            'flake8',
            'sphinx',
            'sphinx-autobuild',
        ),
    },
    classifiers=(
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
    )
)
